<?php

/**
 * Parse a given KML string.
 * 
 * @todo: Add KMZ support (grab code from parser_csv).
 *
 * @param $url 
 *   URL that points to a valid KML file.
 * @return 
 *   a feed object with parsed and normalized data
 */
function _parser_kml_parse($url) {
  // Create FeedAPI feed object.
  $feed = new stdClass();
  $feed->title = t('Placemarks from !url', array('!url' => $url));
  $feed->description = '';
  $feed->options->link = $url;
  $feed->items = array();

  // Download and parse document @ URL
  $kml_string = parser_kml_download($url);  
  $reader = new XMLReader();
  $reader->xml($kml_string);
  $handler = new parserKMLPlacemarkHandler();
  parser_kml_get_elements($reader, $handler, "Placemark", $feed->items);
  $reader->close();
  return $feed;
}

/**
 * @param $reader
 *   An instance of XMLReader.
 * @param $handler
 *   An implementation of the parserKMLHandler interface.
 * @param $element_name
 *   The name of the element to parse.
 * @param $items
 *   The array to add parsed elements to.
 * @param $limit;
 *   The max number of elements to parse.
 */
function parser_kml_get_elements(XMLReader $reader, $handler, $element_name, &$items, $limit = false) {
  $field = null;
  while($reader->read()) {
    if ($reader->nodeType != XMLReader::END_ELEMENT) {
      if ($reader->localName == $element_name) {
        $handler->reset($reader);
      }
      else {
        $field = &$handler->elementProcess($reader, $field);
      }
    }
    elseif ($reader->localName == $element_name && isset($handler)) {
      $item = $handler->elementEnd();
 
      // If this is a place mark, attach namespaces used in placemark and
      // reset namespaces.
      if ($element_name == 'Placemark') {
        $item->namespaces = parserKMLNamespaces::getNameSpaces();
        parserKMLNamespaces::resetNameSpaces();
      }
      $items[] = $item;
      if (is_numeric($limit) && count($items) >= $limit)  {
        return;
      }
    }
  }
}

interface parserKMLHandler  {
  /**
   * @param $reader
   *   An instance of XMLReader
   */
  public function reset($reader);
  
  /**
   * @param $reader
   *   An instance of XMLReader
   * @param $field
   *   A reference to a item on the internal feed item
   * @return
   *   A reference to a item on the internal feed item
   */
  public function &elementProcess(XMLReader $reader, &$field);
  
  /**
   * @return
   *   A parsed and processed element to at the the items array.
   */
  public function elementEnd();
}

class parserKMLPlacemarkHandler implements parserKMLHandler {
  private $item;
  public $depth;

  function reset($reader) {
    $this->depth = $reader->depth;

    $this->item = new stdClass();
    $this->item->title = t('No title');
    $this->item->description = '';
    $this->item->options = new stdClass();
    $this->item->options->kml_type = 'placemark';

    if ($reader->moveToAttribute('id')) {
      $item->options->placemark = $reader->value;
    }
    return;
  }

  /**
   * @todo: 
   *   * comment
   *   * consolidate null pointer referencing and usage of readInnerXml() for simple 
   *     plain text retrievals.
   * 
   * @param $reader XMLreader object
   * @param $field, a reference to the field on the current item that we're populating.
   * @return reference to current field.
   */
  function &elementProcess(XMLReader $reader, &$field) {
    switch ($reader->localName) {
      case 'name':
        return $this->item->title;
      case 'description':
        return $this->item->description;
      case 'link':
        $this->item->options->original_url = $reader->getAttribute('href');
        return null;
      case 'Metadata':
        $this->item->options->metadata = '';
        $handler = new parserKMLMetadataHandler();
        $handler->reset($reader);
        $data = array();
        parser_kml_get_elements($reader, $handler, 'Metadata', $data, 1);
        $this->item->options->metadata = array_pop($data);
        return null;
      case 'icon':
        $this->item->options->icons[] = $reader->readString();
        return null;
      case 'coordinates':
        // Normalize to FeedAPI convention.
        list($lat, $lon) = explode(',', $reader->readInnerXml());
        $this->item->options->location->latitude[0] = $lat;
        $this->item->options->location->longitude[0] = $lon;
        return null;
      default:
        if ($field !== null) {
          if ($reader->nodeType === XMLReader::TEXT || $reader->nodeType === XMLReader::CDATA) {
            $field = $reader->value;
          }
          unset($field);
        }
        break;
    }
    return null;
  }

  function elementEnd() {
    // Generate GUID from URL + hash of serialized placemark. Let's not use the position of this
    // placemark in feed, uniqueness of item doesn't change w/ position in KML feed.
    $this->item->options->guid = hash('md5', serialize($this->item));
    return $this->item;
  }
}

class parserKMLMetadataHandler implements parserKMLHandler {
  private $item;
  private $elem;
  public $depth;

  function reset($reader) {
    $this->depth = $reader->depth;
    $item = array();
  }

  function &elementProcess(XMLReader $reader, &$field) {
    switch ($reader->nodeType) {
      case XMLReader::ELEMENT:
        $this->elem = $reader->name;
        break;
      case XMLReader::TEXT:
        if ($this->elem) {
          parserKMLNamespaces::registerNamespace($reader, $this->elem);
          $this->item[$this->elem] = $reader->value;
          $this->elem = false;
        }
        break;
    }
    return null;
  }

  function elementEnd() {
    return $this->item;
  }
}

/**
 * Static class used to register namespaces.
 */
class parserKMLNamespaces {
  static private $nameSpaces = array();
  
  /**
   * Register namespace in given $elem.
   *
   * @param XMLReader $reader
   * @param string $elem
   */
  static function registerNamespace($reader, $elem) {
    // Resolve namespaces.
    $e = explode(':', $elem);
    if (count($e) == 2) {
      if ($ns = $reader->lookupNamespace($e[0])) {
        self::$nameSpaces[$e[0]] = $ns;
      }
    }
  }

  /**
   * Returns all nameSpaces registered during parsing.
   * 
   * @return Array where keys are namespace prefixes and values are
   *   Base URI's
   */
  static function getNamespaces() {
    return self::$nameSpaces;
  }
  
  /**
   * Resets all namespaces.
   */
  static function resetNamespaces() {
    self::$nameSpaces = array();
  }
}

/**
 * Download file from URL. Unzip file if zipped.
 * @todo: 
 *   This is copied rom parser_csv. Move to feedapi and consolidate.
 */
function parser_kml_download($url) {
  
  // If data is zipped, download and store it.
  if (strpos(substr($url, strlen($url) - 4), '.kmz') !== FALSE || strpos(substr($url, strlen($url) - 8), '.kmztest') !== FALSE) {
    // Download file and unzip it before reading it.
    $result = parser_kml_unzip($url);
  }
  else {
    $result = drupal_http_request($url);
  }
  return $result->data;
}

/**
 * Download unzip and load a zip file at the given URL.
 * @param $url
 *   URL that points to a valid zip file.
 * @return 
 *   return value of drupal_http_request() data element unzipped.
 */
function parser_kml_unzip($url) {
  
  // @todo: don't pipe through PHP, but copy to disk first.
  $result = drupal_http_request($url);
  
  if ($result->error) {
    watchdog('parser_cvs', 'Couldn\'t download file from !url', array('!url' => $url), WATCHDOG_ERROR);
    drupal_set_message(t('Could not download file from !url', array('!url' => $url)), 'error');
    return FALSE;
  }

  // Write file to disk
  $tmp_file = file_directory_temp() .'/unzip_'. hash('md5', $url) .'.zip';
  $tmp_fp = fopen($tmp_file, 'w');
  fwrite($tmp_fp, $result->data);
  fclose($tmp_fp);

  // Unzip it
  $command = 'unzip -o '. $tmp_file .' -d '. file_directory_temp() .' | grep inflating';
  $output = shell_exec($command);
  $matches = array();
  if (preg_match('/inflating: (.*\.kml)/', $output, $matches)) {
    $unzipped_file = $matches[1];
  }

  // Read unzipped data 
  $unzipped_fp = fopen($unzipped_file, 'r');
  $data = fread($unzipped_fp, filesize($unzipped_file));
  fclose($unzipped_fp);
  
  // Stick it onto $result->data or throw error.
  if ($data) {
    $result->data = $data;
    return $result;
  }
  else {
    watchdog('parser_cvs', 'Could not read unzipped file !file', array('!file' => $unzipped_file), WATCHDOG_ERROR);
    drupal_set_message(t('Could not download unzipped file !file', array('!file' => $unzipped_file)), 'error');
  }
  return FALSE;
}